import sys

from jupyterlab_server.translation_utils import (
    TranslationBundle,
    _main,
    get_installed_packages_locale,
    get_language_packs,
    translator,
)


def test_transutils_main():
    sys.argv = ["", "get_language_packs"]
    _main()
    sys.argv = [""]


def test_get_installed_packages_locale(jp_environ):
    get_installed_packages_locale("es_co")


def test_get_language_packs(jp_environ):
    get_language_packs("es_co")


def test_translation_bundle():
    bundle = TranslationBundle("foo", "bar")
    bundle.update_locale("fizz")
    bundle.gettext("hi")
    bundle.ngettext("hi", "his", 1)
    bundle.npgettext("foo", "bar", "bars", 2)
    bundle.pgettext("foo", "bar")


def test_translator():
    t = translator()
    t.load("foo")
    t.normalize_domain("bar")
    t.set_locale("fizz")
    t.translate_schema({})



"""Testing utils."""
import json
import os
import sys
from http.cookies import SimpleCookie
from pathlib import Path
from typing import Optional
from urllib.parse import parse_qs, urlparse

import tornado.httpclient
import tornado.web
from openapi_core.spec.paths import Spec
from openapi_core.validation.request import openapi_request_validator
from openapi_core.validation.request.datatypes import RequestParameters
from openapi_core.validation.response import openapi_response_validator
from tornado.httpclient import HTTPRequest, HTTPResponse
from werkzeug.datastructures import Headers, ImmutableMultiDict

from jupyterlab_server.spec import get_openapi_spec

HERE = Path(os.path.dirname(__file__)).resolve()

with open(HERE / "test_data" / "app-settings" / "overrides.json", encoding="utf-8") as fpt:
    big_unicode_string = json.load(fpt)["@jupyterlab/unicode-extension:plugin"]["comment"]


class TornadoOpenAPIRequest:
    """
    Converts a torando request to an OpenAPI one
    """

    def __init__(self, request: HTTPRequest, spec: Spec):
        """Initialize the request."""
        self.request = request
        self.spec = spec
        if request.url is None:
            raise RuntimeError("Request URL is missing")
        self._url_parsed = urlparse(request.url)

        cookie: SimpleCookie = SimpleCookie()
        cookie.load(request.headers.get("Set-Cookie", ""))
        cookies = {}
        for key, morsel in cookie.items():
            cookies[key] = morsel.value

        # extract the path
        o = urlparse(request.url)

        # gets deduced by path finder against spec
        path: dict = {}

        self.parameters = RequestParameters(
            query=ImmutableMultiDict(parse_qs(o.query)),
            header=Headers(dict(request.headers)),
            cookie=ImmutableMultiDict(cookies),
            path=path,
        )

    @property
    def host_url(self) -> str:
        url = self.request.url
        return url[: url.index('/lab')]

    @property
    def path(self) -> str:
        # extract the best matching url
        # work around lack of support for path parameters which can contain slashes
        # https://github.com/OAI/OpenAPI-Specification/issues/892
        url = None
        o = urlparse(self.request.url)
        for path in self.spec["paths"]:
            if url:
                continue
            has_arg = "{" in path
            if has_arg:
                path = path[: path.index("{")]
            if path in o.path:
                u = o.path[o.path.index(path) :]
                if not has_arg and len(u) == len(path):
                    url = u
                if has_arg and not u.endswith("/"):
                    url = u[: len(path)] + r"foo"

        if url is None:
            raise ValueError(f"Could not find matching pattern for {o.path}")
        return url

    @property
    def method(self) -> str:
        method = self.request.method
        return method and method.lower() or ""

    @property
    def body(self) -> Optional[str]:
        if not isinstance(self.request.body, bytes):
            raise AssertionError('Request body is invalid')
        return self.request.body.decode("utf-8")

    @property
    def mimetype(self) -> str:
        # Order matters because all tornado requests
        # include Accept */* which does not necessarily match the content type
        request = self.request
        return (
            request.headers.get("Content-Type")
            or request.headers.get("Accept")
            or "application/json"
        )


class TornadoOpenAPIResponse:
    """A tornado open API response."""

    def __init__(self, response: HTTPResponse):
        """Initialize the response."""
        self.response = response

    @property
    def data(self) -> str:
        if not isinstance(self.response.body, bytes):
            raise AssertionError('Response body is invalid')
        return self.response.body.decode("utf-8")

    @property
    def status_code(self) -> int:
        return int(self.response.code)

    @property
    def mimetype(self) -> str:
        return str(self.response.headers.get("Content-Type", "application/json"))

    @property
    def headers(self) -> Headers:
        return Headers(dict(self.response.headers))


def validate_request(response):
    """Validate an API request"""
    openapi_spec = get_openapi_spec()

    request = TornadoOpenAPIRequest(response.request, openapi_spec)
    result = openapi_request_validator.validate(openapi_spec, request)
    result.raise_for_errors()

    response = TornadoOpenAPIResponse(response)
    result2 = openapi_response_validator.validate(openapi_spec, request, response)
    result2.raise_for_errors()


def maybe_patch_ioloop():
    """a windows 3.8+ patch for the asyncio loop"""
    if sys.platform.startswith("win") and tornado.version_info < (6, 1):
        if sys.version_info >= (3, 8):

            try:
                from asyncio import WindowsProactorEventLoopPolicy, WindowsSelectorEventLoopPolicy
            except ImportError:
                pass
                # not affected
            else:
                from asyncio import get_event_loop_policy, set_event_loop_policy

                if type(get_event_loop_policy()) is WindowsProactorEventLoopPolicy:
                    # WindowsProactorEventLoopPolicy is not compatible with tornado 6
                    # fallback to the pre-3.8 default of Selector
                    set_event_loop_policy(WindowsSelectorEventLoopPolicy())


def expected_http_error(error, expected_code, expected_message=None):
    """Check that the error matches the expected output error."""
    e = error.value
    if isinstance(e, tornado.web.HTTPError):
        if expected_code != e.status_code:
            return False
        if expected_message is not None and expected_message != str(e):
            return False
        return True
    elif any(
        [
            isinstance(e, tornado.httpclient.HTTPClientError),
            isinstance(e, tornado.httpclient.HTTPError),
        ]
    ):
        if expected_code != e.code:
            return False
        if expected_message:
            message = json.loads(e.response.body.decode())["message"]
            if expected_message != message:
                return False
        return True

    return False



"""Basic tests for the lab handlers.
"""

import pytest
import tornado.httpclient

from jupyterlab_server.test_utils import expected_http_error


@pytest.fixture
def notebooks(jp_create_notebook, labserverapp):
    nbpaths = (
        "notebook1.ipynb",
        "jlab_test_notebooks/notebook2.ipynb",
        "jlab_test_notebooks/level2/notebook3.ipynb",
    )
    for nb in nbpaths:
        jp_create_notebook(nb)
    return nbpaths


async def test_lab_handler(notebooks, jp_fetch):
    r = await jp_fetch("lab", "jlab_test_notebooks")
    assert r.code == 200
    # Check that the lab template is loaded
    html = r.body.decode()
    assert "Files" in html
    assert "JupyterLab Server Application" in html


async def test_notebook_handler(notebooks, jp_fetch):
    for nbpath in notebooks:
        r = await jp_fetch("lab", nbpath)
        assert r.code == 200
        # Check that the lab template is loaded
        html = r.body.decode()
        assert "JupyterLab Server Application" in html


async def test_404(notebooks, jp_fetch):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch("foo")
    assert expected_http_error(e, 404)



"""Test the kernels service API."""
import json
import os

import pytest
import tornado.httpclient
from strict_rfc3339 import rfc3339_to_timestamp  # type:ignore

from jupyterlab_server.test_utils import (
    big_unicode_string,
    expected_http_error,
    maybe_patch_ioloop,
    validate_request,
)

maybe_patch_ioloop()


async def test_delete(jp_fetch, labserverapp):
    orig = "f/o/o"
    copy = "baz"
    r = await jp_fetch("lab", "api", "workspaces", orig)
    validate_request(r)
    res = r.body.decode()
    data = json.loads(res)
    data["metadata"]["id"] = copy
    r2 = await jp_fetch("lab", "api", "workspaces", copy, method="PUT", body=json.dumps(data))
    assert r2.code == 204
    r3 = await jp_fetch(
        "lab",
        "api",
        "workspaces",
        copy,
        method="DELETE",
    )
    assert r3.code == 204
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch(
            "lab",
            "api",
            "workspaces",
            "does_not_exist",
            method="DELETE",
        )
        assert expected_http_error(e, 404)

    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch(
            "lab",
            "api",
            "workspaces",
            "",
            method="DELETE",
        )
        assert expected_http_error(e, 400)


async def test_get_non_existant(jp_fetch, labserverapp):
    id = "foo"

    r = await jp_fetch("lab", "api", "workspaces", id)
    validate_request(r)
    data = json.loads(r.body.decode())

    r2 = await jp_fetch("lab", "api", "workspaces", id, method="PUT", body=json.dumps(data))
    validate_request(r2)

    r3 = await jp_fetch("lab", "api", "workspaces", id)
    validate_request(r3)
    data = json.loads(r3.body.decode())
    first_metadata = data["metadata"]
    first_created = rfc3339_to_timestamp(first_metadata["created"])
    first_modified = rfc3339_to_timestamp(first_metadata["last_modified"])

    r4 = await jp_fetch("lab", "api", "workspaces", id, method="PUT", body=json.dumps(data))
    validate_request(r4)

    r5 = await jp_fetch("lab", "api", "workspaces", id)
    validate_request(r5)
    data = json.loads(r5.body.decode())
    second_metadata = data["metadata"]
    second_created = rfc3339_to_timestamp(second_metadata["created"])
    second_modified = rfc3339_to_timestamp(second_metadata["last_modified"])

    assert first_created <= second_created
    assert first_modified < second_modified


@pytest.mark.skipif(os.name == "nt", reason="Temporal failure on windows")
async def test_get(jp_fetch, labserverapp):
    id = "foo"
    r = await jp_fetch("lab", "api", "workspaces", id)
    validate_request(r)
    data = json.loads(r.body.decode())
    metadata = data["metadata"]
    assert metadata["id"] == id
    assert rfc3339_to_timestamp(metadata["created"])
    assert rfc3339_to_timestamp(metadata["last_modified"])

    r2 = await jp_fetch("lab", "api", "workspaces", id)
    validate_request(r2)
    data = json.loads(r.body.decode())
    assert data["metadata"]["id"] == id


async def test_listing(jp_fetch, labserverapp):
    # ID fields are from workspaces/*.jupyterlab-workspace
    listing = {"foo", "f/o/o/"}
    r = await jp_fetch("lab", "api", "workspaces/")
    validate_request(r)
    res = r.body.decode()
    data = json.loads(res)
    output = set(data["workspaces"]["ids"])
    assert output == listing


async def test_listing_dates(jp_fetch, labserverapp):
    r = await jp_fetch("lab", "api", "workspaces")
    data = json.loads(r.body.decode())
    values = data["workspaces"]["values"]
    times: list = sum(
        ([ws["metadata"].get("last_modified"), ws["metadata"].get("created")] for ws in values), []
    )
    assert None not in times
    [rfc3339_to_timestamp(t) for t in times]


async def test_put(jp_fetch, labserverapp):
    id = "foo"
    r = await jp_fetch("lab", "api", "workspaces", id)
    assert r.code == 200
    res = r.body.decode()
    data = json.loads(res)
    data["metadata"]["big-unicode-string"] = big_unicode_string[::-1]
    r2 = await jp_fetch("lab", "api", "workspaces", id, method="PUT", body=json.dumps(data))
    assert r2.code == 204


async def test_bad_put(jp_fetch, labserverapp):
    orig = "foo"
    copy = "bar"
    r = await jp_fetch("lab", "api", "workspaces", orig)
    assert r.code == 200
    res = r.body.decode()
    data = json.loads(res)
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch("lab", "api", "workspaces", copy, method="PUT", body=json.dumps(data))
    assert expected_http_error(e, 400)


async def test_blank_put(jp_fetch, labserverapp):
    orig = "foo"
    r = await jp_fetch("lab", "api", "workspaces", orig)
    assert r.code == 200
    res = r.body.decode()
    data = json.loads(res)
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch("lab", "api", "workspaces", method="PUT", body=json.dumps(data))
    assert expected_http_error(e, 400)
